from dataclasses import dataclass
from typing import Optional

@dataclass
class RPyFlags:
    typed: bool = False
    fast: bool = False
    no_runtime: bool = False
    shared_runtime: bool = False
    source_refs: bool = False
    compile_time: bool = False
    debug: bool = False
    debug_types: bool = False
    opt_level: int = 1
    # Missing CLI attributes
    explain: bool = False
    verify_cfg: bool = False
    opt: int = 1
    script_type: str = "module"
    profile_path: Optional[str] = None
    backup_studio: bool = False
    show_out: bool = False

# Backwards-compatible alias (many files import CompilerFlags)
CompilerFlags = RPyFlags
