# RPy JSON StdLib
# Optimized wrapper for Roblox HttpService

import _HttpService

def dumps(obj):
    # Luau side will handle table to JSON conversion
    return _HttpService.JSONEncode(obj)

def loads(s):
    # Luau side will handle JSON to table conversion
    return _HttpService.JSONDecode(s)
