import json
import math

def test_debug():
    x = 10
    y = math.sqrt(x)
    data = {"result": y}
    print(json.dumps(data))
    # Intentional error to verify traceback
    return data["undefined_key"]

test_debug()
