"""
sdk/task.py — Luau task library stubs for RPy.
"""

from typing import Any, Callable

def wait(duration: float = ...) -> float: ...
def spawn(callback: Callable, *args: Any) -> Any: ...
def delay(duration: float, callback: Callable, *args: Any) -> Any: ...
def defer(callback: Callable, *args: Any) -> Any: ...
def cancel(thread: Any) -> None: ...
