# RPy Base64 StdLib
# Optimized wrapper for Roblox HttpService

import _HttpService

def b64encode(s):
    if isinstance(s, str):
        return _HttpService.Base64Encode(s)
    # Placeholder for bytes handling if RPy adds it later
    return _HttpService.Base64Encode(str(s))

def b64decode(s):
    return _HttpService.Base64Decode(s)
