"""
sdk/string.py — Luau string library stubs for RPy.
"""

from typing import Any, Optional

def byte(s: str, i: int = 1, j: int = ...) -> Any: ...
def char(*args: int) -> str: ...
def dump(f: Any) -> str: ...
def find(s: str, pattern: str, init: int = 1, plain: bool = False) -> tuple: ...
def format(format_string: str, *args: Any) -> str: ...
def gmatch(s: str, pattern: str) -> Any: ...
def gsub(s: str, pattern: str, repl: Any, n: int = ...) -> tuple[str, int]: ...
def len(s: str) -> int: ...
def lower(s: str) -> str: ...
def match(s: str, pattern: str, init: int = 1) -> Any: ...
def pack(format: str, *args: Any) -> str: ...
def packsize(format: str) -> int: ...
def rep(s: str, n: int) -> str: ...
def reverse(s: str) -> str: ...
def sub(s: str, i: int, j: int = -1) -> str: ...
def unpack(format: str, s: str, pos: int = 1) -> Any: ...
def upper(s: str) -> str: ...
def split(s: str, separator: str) -> list[str]: ...
