# tests/test_phase_a.py

def complex_branching(a, b, c):
    if a:
        x = 1
    elif b:
        x = 2
    else:
        x = 3
    
    # Redundant temp test
    y = x
    z = y
    return z + 1

def nested_if(a, b):
    if a:
        if b:
            return 1
        return 2
    return 3
