# test_explain.py — tests the --explain flag
# Has constant folding, potential DCE, and multiple expressions

def compute(a, b):
    x = a + b
    y = a + b       # duplicate — CSE hit
    dead = x * 0    # dead code if unused
    result = x + y
    return result
