import math as _math

# RPy Math StdLib
# Optimized mapping to Luau math library

pi = 3.141592653589793
e = 2.718281828459045

def sin(x):
    return _math.sin(x)

def cos(x):
    return _math.cos(x)

def tan(x):
    return _math.tan(x)

def asin(x):
    return _math.asin(x)

def acos(x):
    return _math.acos(x)

def atan(x):
    return _math.atan(x)

def atan2(y, x):
    return _math.atan2(y, x)

def sqrt(x):
    return _math.sqrt(x)

def floor(x):
    return _math.floor(x)

def ceil(x):
    return _math.ceil(x)

def pow(x, y):
    return _math.pow(x, y)

def log(x, base=e):
    if base == e:
        return _math.log(x)
    return _math.log(x) / _math.log(base)

def exp(x):
    return _math.exp(x)

def abs(x):
    return _math.abs(x)

def rad(x):
    return _math.rad(x)

def deg(x):
    return _math.deg(x)
