"""
sdk/table.py — Luau table library stubs for RPy.
"""

from typing import Any, Optional, Callable

def insert(t: list, pos_or_value: Any, value: Any = ...) -> None: ...
def remove(t: list, pos: int = ...) -> Any: ...
def sort(t: list, comp: Callable[[Any, Any], bool] = ...) -> None: ...
def concat(t: list, sep: str = "", i: int = 1, j: int = ...) -> str: ...
def pack(*args: Any) -> list: ...
def unpack(t: list, i: int = 1, j: int = ...) -> Any: ...
def find(t: list, value: Any, init: int = 1) -> Optional[int]: ...
def create(count: int, value: Any = ...) -> list: ...
def clear(t: dict | list) -> None: ...
def move(a1: list, f: int, e: int, t: int, a2: list = ...) -> list: ...
def foreachi(t: list, f: Callable[[int, Any], None]) -> None: ...
def foreach(t: dict, f: Callable[[Any, Any], None]) -> None: ...
def getn(t: list) -> int: ...
